# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
Nucleic acid analysis --- :mod:`MDAnalysis.analysis.nuclinfo`
=============================================================

:Author: Elizabeth Denning
:Year: 2011
:Copyright: Lesser GNU Public License v2.1+

The module provides functions to analyze nucleic acid structures, in
particular

- backbone dihedrals,
- chi dihedrals,
- AS or CP phase angles,
- Watson-Crick N1-N3 distances, C2-O2 distances, N6-O4 distances, O6-N4 distances.

For applications of this kind of analysis see
:footcite:p:`Denning2011,Denning2012`.

All functions take a :class:`~MDAnalysis.core.universe.Universe` as an
argument together with further parameters that specify the base or bases in
question. Angles are in degrees. The functions use standard CHARMM names for
nucleic acids and atom names.


.. rubric:: References

.. footbibliography::

Distances
---------

.. autofunction:: wc_pair

.. autofunction:: minor_pair

.. autofunction:: major_pair


Phases
------

.. autofunction:: phase_cp

.. autofunction:: phase_as


Dihedral angles
---------------

.. autofunction:: tors

.. autofunction:: tors_alpha

.. autofunction:: tors_beta

.. autofunction:: tors_gamma

.. autofunction:: tors_delta

.. autofunction:: tors_eps

.. autofunction:: tors_zeta

.. autofunction:: tors_chi

.. autofunction:: hydroxyl

.. autofunction:: pseudo_dihe_baseflip

"""
import numpy as np
from math import pi, sin, cos, sqrt, pow

from MDAnalysis.lib import mdamath


def wc_pair(universe, i, bp, seg1="SYSTEM", seg2="SYSTEM"):
    """Watson-Crick basepair distance for residue `i` with residue `bp`.

    The distance of the nitrogen atoms in a Watson-Crick hydrogen bond is
    computed.

    Parameters
    ----------
    universe : Universe
         :class:`~MDAnalysis.core.universe.Universe` containing the trajectory
    i : int
        resid of the first base
    bp : int
        resid of the second base
    seg1 : str (optional)
        segment id for first base ["SYSTEM"]
    seg2 : str (optional)
        segment id for second base ["SYSTEM"]

    Returns
    -------
    float
        Watson-Crick base pair distance


    Notes
    -----
    If failure occurs be sure to check the segment identification.


    .. versionadded:: 0.7.6
    """
    if universe.select_atoms(" resid {0!s} ".format(i)).resnames[0] in [
        "DC",
        "DT",
        "U",
        "C",
        "T",
        "CYT",
        "THY",
        "URA",
    ]:
        a1, a2 = "N3", "N1"
    if universe.select_atoms(" resid {0!s} ".format(i)).resnames[0] in [
        "DG",
        "DA",
        "A",
        "G",
        "ADE",
        "GUA",
    ]:
        a1, a2 = "N1", "N3"
    wc_dist = universe.select_atoms(
        "(segid {0!s} and resid {1!s} and name {2!s}) "
        "or (segid {3!s} and resid {4!s} and name {5!s}) ".format(
            seg1, i, a1, seg2, bp, a2
        )
    )
    wc = mdamath.norm(wc_dist[0].position - wc_dist[1].position)
    return wc


def minor_pair(universe, i, bp, seg1="SYSTEM", seg2="SYSTEM"):
    """Minor-Groove basepair distance for residue `i` with residue `bp`.

    The distance of the nitrogen and oxygen atoms in a Minor-groove hydrogen
    bond is computed.

    Parameters
    ----------
    universe : Universe
         :class:`~MDAnalysis.core.universe.Universe` containing the trajectory
    i : int
        resid of the first base
    bp : int
        resid of the second base
    seg1 : str (optional)
        segment id for first base ["SYSTEM"]
    seg2 : str (optional)
        segment id for second base ["SYSTEM"]

    Returns
    -------
    float
        Minor groove base pair distance

    Notes
    -----
    If failure occurs be sure to check the segment identification.


    .. versionadded:: 0.7.6
    """
    if universe.select_atoms(" resid {0!s} ".format(i)).resnames[0] in [
        "DC",
        "DT",
        "U",
        "C",
        "T",
        "CYT",
        "THY",
        "URA",
    ]:
        a1, a2 = "O2", "C2"
    if universe.select_atoms(" resid {0!s} ".format(i)).resnames[0] in [
        "DG",
        "DA",
        "A",
        "G",
        "ADE",
        "GUA",
    ]:
        a1, a2 = "C2", "O2"
    c2o2_dist = universe.select_atoms(
        "(segid {0!s} and resid {1!s} and name {2!s}) "
        "or (segid {3!s} and resid {4!s} and name {5!s})".format(
            seg1, i, a1, seg2, bp, a2
        )
    )
    c2o2 = mdamath.norm(c2o2_dist[0].position - c2o2_dist[1].position)
    return c2o2


def major_pair(universe, i, bp, seg1="SYSTEM", seg2="SYSTEM"):
    """Major-Groove basepair distance for residue `i` with residue `bp`.

    The distance of the nitrogen and oxygen atoms in a Major-groove hydrogen
    bond is computed.


    Parameters
    ----------
    universe : Universe
         :class:`~MDAnalysis.core.universe.Universe` containing the trajectory
    i : int
        resid of the first base
    bp : int
        resid of the second base
    seg1 : str (optional)
        segment id for first base ["SYSTEM"]
    seg2 : str (optional)
        segment id for second base ["SYSTEM"]

    Returns
    -------
    float
        Major groove base pair distance

    Notes
    -----
    If failure occurs be sure to check the segment identification.


    .. versionadded:: 0.7.6
    """
    if universe.select_atoms(" resid {0!s} ".format(i)).resnames[0] in [
        "DC",
        "DG",
        "C",
        "G",
        "CYT",
        "GUA",
    ]:
        if universe.select_atoms(" resid {0!s} ".format(i)).resnames[0] in [
            "DC",
            "C",
            "CYT",
        ]:
            a1, a2 = "N4", "O6"
        else:
            a1, a2 = "O6", "N4"
    if universe.select_atoms(" resid {0!s} ".format(i)).resnames[0] in [
        "DT",
        "DA",
        "A",
        "T",
        "U",
        "ADE",
        "THY",
        "URA",
    ]:
        if universe.select_atoms(" resid {0!s} ".format(i)).resnames[0] in [
            "DT",
            "T",
            "THY",
            "U",
            "URA",
        ]:
            a1, a2 = "O4", "N6"
        else:
            a1, a2 = "N6", "O4"
    no_dist = universe.select_atoms(
        "(segid {0!s} and resid {1!s} and name {2!s}) "
        "or (segid {3!s} and resid {4!s} and name {5!s}) ".format(
            seg1, i, a1, seg2, bp, a2
        )
    )
    major = mdamath.norm(no_dist[0].position - no_dist[1].position)
    return major


def phase_cp(universe, seg, i):
    """Pseudo-angle describing the phase of the ribose pucker for residue `i` using the CP method.

    The angle is computed by the positions of atoms in the ribose ring.


    Parameters
    ----------
    universe : Universe
         :class:`~MDAnalysis.core.universe.Universe` containing the trajectory
    seg : str
        segment id for base
    i : int
        resid of the first base

    Returns
    -------
    float
        phase angle in degrees


    .. versionadded:: 0.7.6
    """
    atom1 = universe.select_atoms(" atom {0!s} {1!s} O4' ".format(seg, i))
    atom2 = universe.select_atoms(" atom {0!s} {1!s} C1' ".format(seg, i))
    atom3 = universe.select_atoms(" atom {0!s} {1!s} C2' ".format(seg, i))
    atom4 = universe.select_atoms(" atom {0!s} {1!s} C3' ".format(seg, i))
    atom5 = universe.select_atoms(" atom {0!s} {1!s} C4' ".format(seg, i))

    data1 = atom1.positions
    data2 = atom2.positions
    data3 = atom3.positions
    data4 = atom4.positions
    data5 = atom5.positions

    r0 = (data1 + data2 + data3 + data4 + data5) * (1.0 / 5.0)
    r1 = data1 - r0
    r2 = data2 - r0
    r3 = data3 - r0
    r4 = data4 - r0
    r5 = data5 - r0

    R1 = (
        (r1 * sin(2 * pi * 0.0 / 5.0))
        + (r2 * sin(2 * pi * 1.0 / 5.0))
        + (r3 * sin(2 * pi * 2.0 / 5.0))
        + (r4 * sin(2 * pi * 3.0 / 5.0))
        + (r5 * sin(2 * pi * 4.0 / 5.0))
    )

    R2 = (
        (r1 * cos(2 * pi * 0.0 / 5.0))
        + (r2 * cos(2 * pi * 1.0 / 5.0))
        + (r3 * cos(2 * pi * 2.0 / 5.0))
        + (r4 * cos(2 * pi * 3.0 / 5.0))
        + (r5 * cos(2 * pi * 4.0 / 5.0))
    )

    x = np.cross(R1[0], R2[0])
    n = x / sqrt(pow(x[0], 2) + pow(x[1], 2) + pow(x[2], 2))

    r1_d = np.dot(r1, n)
    r2_d = np.dot(r2, n)
    r3_d = np.dot(r3, n)
    r4_d = np.dot(r4, n)
    r5_d = np.dot(r5, n)

    D = (
        (
            (r1_d * sin(4 * pi * 0.0 / 5.0))
            + (r2_d * sin(4 * pi * 1.0 / 5.0))
            + (r3_d * sin(4 * pi * 2.0 / 5.0))
            + (r4_d * sin(4 * pi * 3.0 / 5.0))
            + (r5_d * sin(4 * pi * 4.0 / 5.0))
        )
        * -1
        * sqrt(2.0 / 5.0)
    )

    C = (
        (r1_d * cos(4 * pi * 0.0 / 5.0))
        + (r2_d * cos(4 * pi * 1.0 / 5.0))
        + (r3_d * cos(4 * pi * 2.0 / 5.0))
        + (r4_d * cos(4 * pi * 3.0 / 5.0))
        + (r5_d * cos(4 * pi * 4.0 / 5.0))
    ) * sqrt(2.0 / 5.0)

    phase_ang = (np.arctan2(D, C) + (pi / 2.0)) * 180.0 / pi
    return phase_ang % 360


def phase_as(universe, seg, i):
    """Pseudo-angle describing the phase of the ribose pucker for residue `i` using the AS method

    The angle is computed by the position vector of atoms in the ribose ring.

    Parameters
    ----------
    universe : Universe
         :class:`~MDAnalysis.core.universe.Universe` containing the trajectory
    seg : str
        segment id for base
    i : int
        resid of the first base

    Returns
    -------
    float
        phase angle in degrees


    .. versionadded:: 0.7.6
    """
    angle1 = universe.select_atoms(
        " atom {0!s} {1!s} C1' ".format(seg, i),
        " atom {0!s} {1!s} C2' ".format(seg, i),
        " atom {0!s} {1!s} C3' ".format(seg, i),
        " atom {0!s} {1!s} C4' ".format(seg, i),
    )

    angle2 = universe.select_atoms(
        " atom {0!s} {1!s} C2' ".format(seg, i),
        " atom {0!s} {1!s} C3' ".format(seg, i),
        " atom {0!s} {1!s} C4' ".format(seg, i),
        " atom {0!s} {1!s} O4' ".format(seg, i),
    )

    angle3 = universe.select_atoms(
        " atom {0!s} {1!s} C3' ".format(seg, i),
        " atom {0!s} {1!s} C4' ".format(seg, i),
        " atom {0!s} {1!s} O4' ".format(seg, i),
        " atom {0!s} {1!s} C1' ".format(seg, i),
    )

    angle4 = universe.select_atoms(
        " atom {0!s} {1!s} C4' ".format(seg, i),
        " atom {0!s} {1!s} O4' ".format(seg, i),
        " atom {0!s} {1!s} C1' ".format(seg, i),
        " atom {0!s} {1!s} C2' ".format(seg, i),
    )

    angle5 = universe.select_atoms(
        " atom {0!s} {1!s} O4' ".format(seg, i),
        " atom {0!s} {1!s} C1' ".format(seg, i),
        " atom {0!s} {1!s} C2' ".format(seg, i),
        " atom {0!s} {1!s} C3' ".format(seg, i),
    )

    data1 = angle1.dihedral.value()
    data2 = angle2.dihedral.value()
    data3 = angle3.dihedral.value()
    data4 = angle4.dihedral.value()
    data5 = angle5.dihedral.value()

    B = (
        (
            (data1 * sin(2 * 2 * pi * (1 - 1.0) / 5.0))
            + (data2 * sin(2 * 2 * pi * (2 - 1.0) / 5.0))
            + (data3 * sin(2 * 2 * pi * (3 - 1.0) / 5.0))
            + (data4 * sin(2 * 2 * pi * (4 - 1.0) / 5.0))
            + (data5 * sin(2 * 2 * pi * (5 - 1.0) / 5.0))
        )
        * -2.0
        / 5.0
    )

    A = (
        (
            (data1 * cos(2 * 2 * pi * (1 - 1.0) / 5.0))
            + (data2 * cos(2 * 2 * pi * (2 - 1.0) / 5.0))
            + (data3 * cos(2 * 2 * pi * (3 - 1.0) / 5.0))
            + (data4 * cos(2 * 2 * pi * (4 - 1.0) / 5.0))
            + (data5 * cos(2 * 2 * pi * (5 - 1.0) / 5.0))
        )
        * 2.0
        / 5.0
    )

    phase_ang = np.arctan2(B, A) * 180.0 / pi
    return phase_ang % 360


def tors(universe, seg, i):
    """Calculation of nucleic backbone dihedral angles.

    The dihedral angles are alpha, beta, gamma, delta, epsilon, zeta, chi.

    The dihedral is computed based on position of atoms for resid `i`.

    Parameters
    ----------
    universe : Universe
         :class:`~MDAnalysis.core.universe.Universe` containing the trajectory
    seg : str
        segment id for base
    i : int
        resid of the first base

    Returns
    -------
    [alpha, beta, gamma, delta, epsilon, zeta, chi] : list of floats
        torsion angles in degrees

    Notes
    -----
    If failure occurs be sure to check the segment identification.


    .. versionadded:: 0.7.6

    """
    a = universe.select_atoms(
        " atom {0!s} {1!s} O3' ".format(seg, i - 1),
        " atom {0!s} {1!s} P  ".format(seg, i),
        " atom {0!s} {1!s} O5' ".format(seg, i),
        " atom {0!s} {1!s} C5' ".format(seg, i),
    )

    b = universe.select_atoms(
        " atom {0!s} {1!s} P    ".format(seg, i),
        " atom {0!s} {1!s} O5' ".format(seg, i),
        " atom {0!s} {1!s} C5' ".format(seg, i),
        " atom {0!s} {1!s} C4' ".format(seg, i),
    )

    g = universe.select_atoms(
        " atom {0!s} {1!s} O5' ".format(seg, i),
        " atom {0!s} {1!s} C5' ".format(seg, i),
        " atom {0!s} {1!s} C4' ".format(seg, i),
        " atom {0!s} {1!s} C3' ".format(seg, i),
    )

    d = universe.select_atoms(
        " atom {0!s} {1!s} C5' ".format(seg, i),
        " atom {0!s} {1!s} C4' ".format(seg, i),
        " atom {0!s} {1!s} C3' ".format(seg, i),
        " atom {0!s} {1!s} O3' ".format(seg, i),
    )

    e = universe.select_atoms(
        " atom {0!s} {1!s} C4' ".format(seg, i),
        " atom {0!s} {1!s} C3' ".format(seg, i),
        " atom {0!s} {1!s} O3' ".format(seg, i),
        " atom {0!s} {1!s} P    ".format(seg, i + 1),
    )

    z = universe.select_atoms(
        " atom {0!s} {1!s} C3' ".format(seg, i),
        " atom {0!s} {1!s} O3' ".format(seg, i),
        " atom {0!s} {1!s} P    ".format(seg, i + 1),
        " atom {0!s} {1!s} O5' ".format(seg, i + 1),
    )
    c = universe.select_atoms(
        " atom {0!s} {1!s} O4' ".format(seg, i),
        " atom {0!s} {1!s} C1' ".format(seg, i),
        " atom {0!s} {1!s} N9 ".format(seg, i),
        " atom {0!s} {1!s} C4  ".format(seg, i),
    )
    if len(c) < 4:
        c = universe.select_atoms(
            " atom {0!s} {1!s} O4' ".format(seg, i),
            " atom {0!s} {1!s} C1' ".format(seg, i),
            " atom {0!s} {1!s} N1 ".format(seg, i),
            " atom {0!s} {1!s} C2  ".format(seg, i),
        )

    alpha = a.dihedral.value() % 360
    beta = b.dihedral.value() % 360
    gamma = g.dihedral.value() % 360
    delta = d.dihedral.value() % 360
    epsilon = e.dihedral.value() % 360
    zeta = z.dihedral.value() % 360
    chi = c.dihedral.value() % 360

    return [alpha, beta, gamma, delta, epsilon, zeta, chi]


def tors_alpha(universe, seg, i):
    """alpha backbone dihedral

    The dihedral is computed based on position atoms for resid `i`.

    Parameters
    ----------
    universe : Universe
         :class:`~MDAnalysis.core.universe.Universe` containing the trajectory
    seg : str
        segment id for base
    i : int
        resid of the first base

    Returns
    -------
    alpha : float
        torsion angle in degrees


    .. versionadded:: 0.7.6
    """
    a = universe.select_atoms(
        " atom {0!s} {1!s} O3' ".format(seg, i - 1),
        " atom {0!s} {1!s} P  ".format(seg, i),
        " atom {0!s} {1!s} O5' ".format(seg, i),
        " atom {0!s} {1!s} C5' ".format(seg, i),
    )
    alpha = a.dihedral.value() % 360
    return alpha


def tors_beta(universe, seg, i):
    """beta  backbone dihedral

    The dihedral is computed based on position atoms for resid `i`.

    Parameters
    ----------
    universe : Universe
         :class:`~MDAnalysis.core.universe.Universe` containing the trajectory
    seg : str
        segment id for base
    i : int
        resid of the first base

    Returns
    -------
    beta : float
        torsion angle in degrees


    .. versionadded:: 0.7.6
    """
    b = universe.select_atoms(
        " atom {0!s} {1!s} P    ".format(seg, i),
        " atom {0!s} {1!s} O5' ".format(seg, i),
        " atom {0!s} {1!s} C5' ".format(seg, i),
        " atom {0!s} {1!s} C4' ".format(seg, i),
    )
    beta = b.dihedral.value() % 360
    return beta


def tors_gamma(universe, seg, i):
    """Gamma backbone dihedral

    The dihedral is computed based on position atoms for resid `i`.

    Parameters
    ----------
    universe : Universe
         :class:`~MDAnalysis.core.universe.Universe` containing the trajectory
    seg : str
        segment id for base
    i : int
        resid of the first base

    Returns
    -------
    gamma : float
        torsion angle in degrees


    .. versionadded:: 0.7.6
    """
    g = universe.select_atoms(
        " atom {0!s} {1!s} O5' ".format(seg, i),
        " atom {0!s} {1!s} C5' ".format(seg, i),
        " atom {0!s} {1!s} C4' ".format(seg, i),
        " atom {0!s} {1!s} C3' ".format(seg, i),
    )
    gamma = g.dihedral.value() % 360
    return gamma


def tors_delta(universe, seg, i):
    """delta backbone dihedral

    The dihedral is computed based on position atoms for resid `i`.

    Parameters
    ----------
    universe : Universe
         :class:`~MDAnalysis.core.universe.Universe` containing the trajectory
    seg : str
        segment id for base
    i : int
        resid of the first base

    Returns
    -------
    delta : float
        torsion angle in degrees


    .. versionadded:: 0.7.6
    """
    d = universe.select_atoms(
        " atom {0!s} {1!s} C5' ".format(seg, i),
        " atom {0!s} {1!s} C4' ".format(seg, i),
        " atom {0!s} {1!s} C3' ".format(seg, i),
        " atom {0!s} {1!s} O3' ".format(seg, i),
    )
    delta = d.dihedral.value() % 360
    return delta


def tors_eps(universe, seg, i):
    """Epsilon backbone dihedral

    The dihedral is computed based on position atoms for resid `i`.

    Parameters
    ----------
    universe : Universe
         :class:`~MDAnalysis.core.universe.Universe` containing the trajectory
    seg : str
        segment id for base
    i : int
        resid of the first base

    Returns
    -------
    epsilon : float
        torsion angle in degrees


    .. versionadded:: 0.7.6
    """
    e = universe.select_atoms(
        " atom {0!s} {1!s} C4' ".format(seg, i),
        " atom {0!s} {1!s} C3' ".format(seg, i),
        " atom {0!s} {1!s} O3' ".format(seg, i),
        " atom {0!s} {1!s} P    ".format(seg, i + 1),
    )
    epsilon = e.dihedral.value() % 360
    return epsilon


def tors_zeta(universe, seg, i):
    """Zeta backbone dihedral

    The dihedral is computed based on position atoms for resid `i`.

    Parameters
    ----------
    universe : Universe
         :class:`~MDAnalysis.core.universe.Universe` containing the trajectory
    seg : str
        segment id for base
    i : int
        resid of the first base

    Returns
    -------
    zeta : float
        torsion angle in degrees


    .. versionadded:: 0.7.6
    """
    z = universe.select_atoms(
        " atom {0!s} {1!s} C3' ".format(seg, i),
        " atom {0!s} {1!s} O3' ".format(seg, i),
        " atom {0!s} {1!s} P    ".format(seg, i + 1),
        " atom {0!s} {1!s} O5' ".format(seg, i + 1),
    )
    zeta = z.dihedral.value() % 360
    return zeta


def tors_chi(universe, seg, i):
    """chi nucleic acid dihedral

     The dihedral is computed based on position atoms for resid `i`.

    Parameters
    ----------
    universe : Universe
         :class:`~MDAnalysis.core.universe.Universe` containing the trajectory
    seg : str
        segment id for base
    i : int
        resid of the first base

    Returns
    -------
    chi : float
        torsion angle in degrees


    .. versionadded:: 0.7.6
    """
    c = universe.select_atoms(
        " atom {0!s} {1!s} O4' ".format(seg, i),
        " atom {0!s} {1!s} C1' ".format(seg, i),
        " atom {0!s} {1!s} N9 ".format(seg, i),
        " atom {0!s} {1!s} C4  ".format(seg, i),
    )
    if len(c) < 4:
        c = universe.select_atoms(
            " atom {0!s} {1!s} O4' ".format(seg, i),
            " atom {0!s} {1!s} C1' ".format(seg, i),
            " atom {0!s} {1!s} N1 ".format(seg, i),
            " atom {0!s} {1!s} C2  ".format(seg, i),
        )
    chi = c.dihedral.value() % 360
    return chi


def hydroxyl(universe, seg, i):
    """2-hydroxyl dihedral. Useful only for RNA calculations.

     .. Note:: This dihedral calculation will only work if using atom
               names as documented by charmm force field parameters,
               namely "C1', C2', O2', H2'".

    Parameters
    ----------
    universe : Universe
         :class:`~MDAnalysis.core.universe.Universe` containing the trajectory
    seg : str
        segment id for base
    i : int
        resid of the first base

    Returns
    -------
    hydroxyl_angle : float
        torsion angle in degrees


    .. versionadded:: 0.7.6

    """
    h = universe.select_atoms(
        "atom {0!s} {1!s} C1'".format(seg, i),
        "atom {0!s} {1!s} C2'".format(seg, i),
        "atom {0!s} {1!s} O2'".format(seg, i),
        "atom {0!s} {1!s} H2'".format(seg, i),
    )
    try:
        hydr = h.dihedral.value() % 360
    except ValueError:
        errmsg = (
            f"Resid {i} does not contain atoms C1', C2', O2', H2' but "
            f"atoms {list(h.atoms)}"
        )
        raise ValueError(errmsg) from None

    return hydr


def pseudo_dihe_baseflip(
    universe, bp1, bp2, i, seg1="SYSTEM", seg2="SYSTEM", seg3="SYSTEM"
):
    """pseudo dihedral for flipped bases. Useful only for nucleic acid base flipping

    The dihedral is computed based on position atoms for resid `i`

    .. Note:: This dihedral calculation will only work if using atom names as
              documented by charmm force field parameters.

    Parameters
    ----------
    universe : Universe
        :class:`~MDAnalysis.core.universe.Universe` containing the
        trajectory
    bp1 : int
        resid that base pairs with `bp2`
    bp2 : int
        resid below the base that flips
    i : int
        resid of the base that flips
    segid1 : str (optional)
        segid of resid base pairing with `bp2`
    segid2 : str (optional)
        segid, same as that of segid of flipping resid `i`
    segid3 : str (optional)
        segid of resid `i` that flips

    Returns
    -------
    float
          pseudo dihedral angle in degrees


    .. versionadded:: 0.8.0
    """
    bf1 = universe.select_atoms(
        " ( segid {0!s} and resid {1!s} and nucleicbase ) "
        "or ( segid {2!s} and resid {3!s} and nucleicbase ) ".format(
            seg1, bp1, seg2, bp2
        )
    )
    bf4 = universe.select_atoms(
        "(segid {0!s} and resid {1!s} and nucleicbase) ".format(seg3, i)
    )
    bf2 = universe.select_atoms(
        "(segid {0!s} and resid {1!s} and nucleicsugar) ".format(seg2, bp2)
    )
    bf3 = universe.select_atoms(
        "(segid {0!s} and resid {1!s} and nucleicsugar) ".format(seg3, i)
    )
    x = [
        bf1.center_of_mass(),
        bf2.center_of_mass(),
        bf3.center_of_mass(),
        bf4.center_of_mass(),
    ]
    pseudo = mdamath.dihedral(x[0] - x[1], x[1] - x[2], x[2] - x[3])
    pseudo = np.rad2deg(pseudo) % 360
    return pseudo



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#


"""XYZ trajectory reader --- :mod:`MDAnalysis.coordinates.XYZ`
==============================================================

The :ref:`XYZ format <xyz-format>` is a loosely defined, simple
coordinate trajectory format. The implemented format definition was
taken from the `VMD xyzplugin`_ and is therefore compatible with VMD.

Note the following:

* Comments are not allowed in the XYZ file (we neither read nor write
  them to remain compatible with VMD).
* The atom name (first column) is ignored during reading.
* The coordinates are assumed to be space-delimited rather than fixed
  width (this may cause issues - see below).
* All fields to the right of the z-coordinate are ignored.
* The unitcell information is all zeros since this is not recorded in
  the XYZ format.

.. rubric:: Units

* Coordinates are in Angstroms.
* The length of a timestep can be set by passing the *dt* argument,
  it's assumed to be in ps (default: 1 ps).

There appears to be no rigid format definition so it is likely users
will need to tweak this class.

.. _xyz-format:

XYZ File format
---------------

Definition used by the :class:`XYZReader` and :class:`XYZWriter` (and
the `VMD xyzplugin`_ from whence the definition was taken)::

    [ comment line            ] !! NOT IMPLEMENTED !! DO NOT INCLUDE
    [ N                       ] # of atoms, required by this xyz reader plugin  line 1
    [ molecule name           ] name of molecule (can be blank)                 line 2
    atom1 x y z [optional data] atom name followed by xyz coords                line 3
    atom2 x y z [ ...         ] and (optionally) other data.
    ...
    atomN x y z [ ...         ]                                                 line N+2


Note
----
* comment lines not implemented (do not include them)
* molecule name: the line is required but the content is ignored
  at the moment
* optional data (after the coordinates) are presently ignored


.. Links
.. _`VMD xyzplugin`: http://www.ks.uiuc.edu/Research/vmd/plugins/molfile/xyzplugin.html

Classes
-------

"""
import itertools
import os
import errno
import numpy as np
import warnings
import logging

logger = logging.getLogger("MDAnalysis.coordinates.XYZ")

from . import base
from .timestep import Timestep
from ..lib import util
from ..lib.util import cached, store_init_arguments
from ..exceptions import NoDataError
from ..version import __version__


class XYZWriter(base.WriterBase):
    """Writes an XYZ file

    The XYZ file format is not formally defined. This writer follows
    the VMD implementation for the molfile `xyzplugin`_.


    Notes
    -----
    By default, the XYZ writer will attempt to use the input
    :class:`~MDAnalysis.core.groups.AtomGroup` or
    :class:`~MDAnalysis.core.universe.Universe` ``elements`` record to assign
    atom names in the XYZ file. If the ``elements`` record is missing, then
    the ``name`` record will be used. In the event that neither of these are
    available, the atoms will all be named ``X``. Please see, the
    `User Guide`_ for more information on how to add topology attributes if
    you wish to add your own elements / atom names to a
    :class:`~MDAnalysis.core.universe.Universe`.


    .. Links

    .. _xyzplugin:
           http://www.ks.uiuc.edu/Research/vmd/plugins/molfile/xyzplugin.html
    .. _User Guide:
           https://userguide.mdanalysis.org/examples/constructing_universe.html#Adding-topology-attributes


    .. versionchanged:: 1.0.0
       Use elements attribute instead of names attribute, if present.
    .. versionchanged:: 2.0.0
       Support for passing timestep to the writer was deprecated in 1.0 and
       has now been removed. As a consequence, custom names can no longer be
       passed to the writer, these should be added to the
       :class:`~MDAnalysis.core.universe.Universe`, or
       :class:`~MDAnalysis.core.groups.AtomGroup` before invoking the writer.
    """

    format = "XYZ"
    multiframe = True
    # these are assumed!
    units = {"time": "ps", "length": "Angstrom"}

    def __init__(
        self,
        filename,
        n_atoms=None,
        convert_units=True,
        remark=None,
        precision=5,
        **kwargs,
    ):
        """Initialize the XYZ trajectory writer

        Parameters
        ----------
        filename: str
            filename of trajectory file. If it ends with "gz" then the file
            will be gzip-compressed; if it ends with "bz2" it will be bzip2
            compressed.
        n_atoms: int (optional)
            Number of atoms in trajectory. By default assume that this is None
            and that this file is used to store several different models
            instead of a single trajectory. If a number is provided each
            written TimeStep has to contain the same number of atoms.
        convert_units : bool (optional)
            convert quantities to default MDAnalysis units of Angstrom upon
            writing  [``True``]
        remark: str (optional)
            single line of text ("molecule name"). By default writes MDAnalysis
            version and frame
        precision: int (optional)
            set precision of saved trjactory to this number of decimal places.

            .. versionadded:: 2.9.0


        .. versionchanged:: 1.0.0
           Removed :code:`default_remark` variable (Issue #2692).
        .. versionchanged:: 2.0.0
           Due to the removal of timestep as an input for writing, the atoms
           parameter is no longer relevant and has been removed. If passing
           an empty universe, please use ``add_TopologyAttr`` to add in the
           required elements or names.
        """
        self.filename = filename
        self.remark = remark
        self.n_atoms = n_atoms
        self.convert_units = convert_units
        self.precision = precision

        # can also be gz, bz2
        self._xyz = util.anyopen(self.filename, "wt")

    def _get_atoms_elements_or_names(self, atoms):
        """Return a list of atom elements (if present) or fallback to atom names"""
        try:
            return atoms.atoms.elements
        except (AttributeError, NoDataError):
            try:
                return atoms.atoms.names
            except (AttributeError, NoDataError):
                wmsg = (
                    "Input AtomGroup or Universe does not have atom "
                    "elements or names attributes, writer will default "
                    "atom names to 'X'"
                )
                warnings.warn(wmsg)
                return itertools.cycle(("X",))

    def close(self):
        """Close the trajectory file and finalize the writing"""
        if self._xyz is not None:
            self._xyz.write("\n")
            self._xyz.close()
        self._xyz = None

    def write(self, obj):
        """Write object `obj` at current trajectory frame to file.

        Atom elements (or names) in the output are taken from the `obj` or
        default to the value of the `atoms` keyword supplied to the
        :class:`XYZWriter` constructor.

        Parameters
        ----------
        obj : Universe or AtomGroup
            The :class:`~MDAnalysis.core.groups.AtomGroup` or
            :class:`~MDAnalysis.core.universe.Universe` to write.


        .. versionchanged:: 2.0.0
           Deprecated support for Timestep argument has now been removed.
           Use AtomGroup or Universe as an input instead.
        """
        # prepare the Timestep and extract atom names if possible
        # (The way it is written it should be possible to write
        # trajectories with frames that differ in atom numbers
        # but this is not tested.)
        try:
            atoms = obj.atoms
        except AttributeError:
            errmsg = "Input obj is neither an AtomGroup or Universe"
            raise TypeError(errmsg) from None
        else:
            if hasattr(obj, "universe"):
                # For AtomGroup and children (Residue, ResidueGroup, Segment)
                ts_full = obj.universe.trajectory.ts
                if ts_full.n_atoms == atoms.n_atoms:
                    ts = ts_full
                else:
                    # Only populate a time step with the selected atoms.
                    ts = ts_full.copy_slice(atoms.indices)
            elif hasattr(obj, "trajectory"):
                # For Universe only --- get everything
                ts = obj.trajectory.ts
            # update atom names
            self.atomnames = self._get_atoms_elements_or_names(atoms)

        self._write_next_frame(ts)

    def _write_next_frame(self, ts=None):
        """
        Write coordinate information in *ts* to the trajectory

        .. versionchanged:: 1.0.0
           Print out :code:`remark` if present, otherwise use generic one
           (Issue #2692).
           Renamed from `write_next_timestep` to `_write_next_frame`.
        """
        if ts is None:
            if not hasattr(self, "ts"):
                raise NoDataError(
                    "XYZWriter: no coordinate data to write to "
                    "trajectory file"
                )
            else:
                ts = self.ts

        if self.n_atoms is not None:
            if self.n_atoms != ts.n_atoms:
                raise ValueError(
                    "n_atoms keyword was specified indicating "
                    "that this should be a trajectory of the "
                    "same model. But the provided TimeStep has a "
                    "different number ({}) then expected ({})"
                    "".format(ts.n_atoms, self.n_atoms)
                )
        else:
            if (
                not isinstance(self.atomnames, itertools.cycle)
                and len(self.atomnames) != ts.n_atoms
            ):
                logger.info(
                    "Trying to write a TimeStep with unknown atoms. "
                    'Expected {} atoms, got {}. Try using "write" if you are '
                    'using "_write_next_frame" directly'.format(
                        len(self.atomnames), ts.n_atoms
                    )
                )
                self.atomnames = np.array([self.atomnames[0]] * ts.n_atoms)

        if self.convert_units:
            coordinates = self.convert_pos_to_native(
                ts.positions, inplace=False
            )
        else:
            coordinates = ts.positions

        # Write number of atoms
        self._xyz.write("{0:d}\n".format(ts.n_atoms))

        # Write remark
        if self.remark is None:
            remark = (
                "frame {} | Written by MDAnalysis {} (release {})\n".format(
                    ts.frame, self.__class__.__name__, __version__
                )
            )

            self._xyz.write(remark)
        else:
            self._xyz.write(self.remark.strip() + "\n")

        # Write content
        for atom, (x, y, z) in zip(self.atomnames, coordinates):
            self._xyz.write(
                "{0!s:>8}  {1:10.{p}f} {2:10.{p}f} {3:10.{p}f}\n"
                "".format(atom, x, y, z, p=self.precision)
            )


class XYZReader(base.ReaderBase):
    """Reads from an XYZ file

    :Data:
        :attr:`ts`
          Timestep object containing coordinates of current frame

    :Methods:
        ``len(xyz)``
          return number of frames in xyz
        ``for ts in xyz:``
          iterate through trajectory

    .. Note: this can read both compressed (foo.xyz) and compressed
          (foo.xyz.bz2 or foo.xyz.gz) files; uncompression is handled
          on the fly and also reads streams via
          :class:`~MDAnalysis.lib.util.NamedStream`.

    The XYZ file format follows VMD's xyzplugin_ and is also described
    under :ref:`XYZ format <xyz-format>`.

    .. versionchanged:: 0.11.0
       Frames now 0-based instead of 1-based. Added *dt* and
       *time_offset* keywords (passed to :class:`Timestep`)
    """

    # Phil Fowler:
    # Validation: the geometric centre of 1284 atoms was calculated over
    # 500 frames using both MDAnalysis and a VMD Tcl script. There was
    # exact agreement (measured to three decimal places). bzipped and
    # gzipped versions of the XYZ file were also tested

    format = "XYZ"
    # these are assumed!
    units = {"time": "ps", "length": "Angstrom"}
    _Timestep = Timestep

    @store_init_arguments
    def __init__(self, filename, **kwargs):
        super(XYZReader, self).__init__(filename, **kwargs)

        # the filename has been parsed to be either be foo.xyz or foo.xyz.bz2 by
        # coordinates::core.py so the last file extension will tell us if it is
        # bzipped or not
        root, ext = os.path.splitext(self.filename)
        self.xyzfile = util.anyopen(self.filename)
        self.compression = ext[1:] if ext[1:] != "xyz" else None
        self._cache = dict()

        self.ts = self._Timestep(self.n_atoms, **self._ts_kwargs)
        # Haven't quite figured out where to start with all the self._reopen()
        # etc.
        # (Also cannot just use seek() or reset() because that would break
        # with urllib2.urlopen() streams)
        self._read_next_timestep()

    @property
    @cached("n_atoms")
    def n_atoms(self):
        """number of atoms in a frame"""
        with util.anyopen(self.filename) as f:
            n = f.readline()
        # need to check type of n
        return int(n)

    @property
    @cached("n_frames")
    def n_frames(self):
        try:
            return self._read_xyz_n_frames()
        except IOError:
            return 0

    def _read_xyz_n_frames(self):
        # the number of lines in the XYZ file will be 2 greater than the
        # number of atoms
        linesPerFrame = self.n_atoms + 2
        counter = 0
        offsets = []

        with util.anyopen(self.filename) as f:
            line = True
            while line:
                if not counter % linesPerFrame:
                    offsets.append(f.tell())
                line = f.readline()
                counter += 1

        # need to check this is an integer!
        n_frames = int(counter / linesPerFrame)
        self._offsets = offsets
        return n_frames

    def _read_frame(self, frame):
        self.xyzfile.seek(self._offsets[frame])
        self.ts.frame = frame - 1  # gets +1'd in next
        return self._read_next_timestep()

    def _read_next_timestep(self, ts=None):
        # check that the timestep object exists
        if ts is None:
            ts = self.ts

        f = self.xyzfile

        try:
            # we assume that there are only two header lines per frame
            f.readline()
            f.readline()
            # convert all entries at the end once for optimal speed
            tmp_buf = []
            for i in range(self.n_atoms):
                tmp_buf.append(f.readline().split()[1:4])
            ts.positions = tmp_buf
            ts.frame += 1
            return ts
        except (ValueError, IndexError) as err:
            raise EOFError(err) from None

    def _reopen(self):
        self.close()
        self.open_trajectory()

    def open_trajectory(self):
        if self.xyzfile is not None:
            raise IOError(
                errno.EALREADY, "XYZ file already opened", self.filename
            )

        self.xyzfile = util.anyopen(self.filename)

        # reset ts
        ts = self.ts
        ts.frame = -1

        return self.xyzfile

    def Writer(self, filename, n_atoms=None, **kwargs):
        """Returns a XYZWriter for *filename* with the same parameters as this
        XYZ.

        Parameters
        ----------
        filename: str
            filename of the output trajectory
        n_atoms: int (optional)
            number of atoms. If none is given use the same number of atoms from
            the reader instance is used
        **kwargs:
            See :class:`XYZWriter` for additional kwargs

        Returns
        -------
        :class:`XYZWriter` (see there for more details)

        See Also
        --------
        :class:`XYZWriter`
        """
        if n_atoms is None:
            n_atoms = self.n_atoms
        return XYZWriter(filename, n_atoms=n_atoms, **kwargs)

    def close(self):
        """Close xyz trajectory file if it was open."""
        if self.xyzfile is None:
            return
        self.xyzfile.close()
        self.xyzfile = None



from libcpp cimport bool
from libc.stdint cimport uint64_t, int64_t
cimport numpy as cnp
cnp.import_array()


cdef class Timestep:

    # number of atoms in the timestep
    cdef uint64_t _n_atoms
    # the frame number of the timestep
    cdef public int64_t  frame


    # Info for NumPy C API
    # NumPy enumerated type, must be one of cnp.NPY_FLOAT32 or cnp.NPY_FLOAT64
    cdef int _typenum
    # shape of particle dependent arrays, eg positions, set to (_n_atoms, 3)
    # on __cinit__
    cdef cnp.npy_intp _particle_dependent_dim[2]

    # indicates whether timestep has particle associated data
    cdef bool _has_positions
    cdef bool _has_velocities
    cdef bool _has_forces


    # tracks whether particle associated data has been allocated correct shape
    cdef bool _positions_alloc
    cdef bool _velocities_alloc
    cdef bool _forces_alloc

    # unitcell and particle dependent data
    # these have to be public for testing
    cdef public cnp.ndarray _unitcell
    cdef public cnp.ndarray _pos
    cdef public cnp.ndarray _velocities
    cdef public cnp.ndarray _forces

    # Python objects associated with the timestep
    # NumPy dtype object
    cdef object _dtype
    # additional data associated with timestep
    cdef public dict data
    # weakref back to reader that fills this timestep
    cdef public object  _reader
    # auxilliary namespace
    cdef public object aux



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""
Guessing unknown Topology information --- :mod:`MDAnalysis.topology.guessers`
=============================================================================

.. deprecated:: 2.8.0
   The :mod:`MDAnalysis.topology.guessers` module will be removed in release 3.0.0.
   It is deprecated in favor of the new Guessers API. See 
   :mod:`MDAnalysis.guesser.default_guesser` for more details.

In general `guess_atom_X` returns the guessed value for a single value,
while `guess_Xs` will work on an array of many atoms.


Example uses of guessers
------------------------

Guessing elements from atom names
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Currently, it is possible to guess elements from atom names using
:func:`guess_atom_element` (or the synonymous :func:`guess_atom_type`). This can
be done in the following manner::

  import MDAnalysis as mda
  from MDAnalysis.topology.guessers import guess_atom_element
  from MDAnalysisTests.datafiles import PRM7

  u = mda.Universe(PRM7)

  print(u.atoms.names[1])  # returns the atom name H1

  element = guess_atom_element(u.atoms.names[1])

  print(element)  # returns element H

In the above example, we take an atom named H1 and use
:func:`guess_atom_element` to guess the element hydrogen (i.e. H). It is
important to note that element guessing is not always accurate. Indeed in cases
where the atom type is not recognised, we may end up with the wrong element.
For example::

  import MDAnalysis as mda
  from MDAnalysis.topology.guessers import guess_atom_element
  from MDAnalysisTests.datafiles import PRM19SBOPC

  u = mda.Universe(PRM19SBOPC)

  print(u.atoms.names[-1])  # returns the atom name EPW

  element = guess_atom_element(u.atoms.names[-1])

  print(element)  # returns element P

Here we find that virtual site atom 'EPW' was given the element P, which
would not be an expected result. We therefore always recommend that users
carefully check the outcomes of any guessers.

In some cases, one may want to guess elements for an entire universe and add
this guess as a topology attribute. This can be done using :func:`guess_types`
in the following manner::

  import MDAnalysis as mda
  from MDAnalysis.topology.guessers import guess_types
  from MDAnalysisTests.datafiles import PRM7

  u = mda.Universe(PRM7)

  guessed_elements = guess_types(u.atoms.names)

  u.add_TopologyAttr('elements', guessed_elements)

  print(u.atoms.elements)  # returns an array of guessed elements

More information on adding topology attributes can found in the `user guide`_.


.. Links

.. _user guide: https://www.mdanalysis.org/UserGuide/examples/constructing_universe.html#Adding-topology-attributes

"""
import numpy as np
import warnings
import re

from ..lib import distances
from MDAnalysis.guesser import tables
from MDAnalysis.lib.util import deprecate


deprecation_msg = (
    "MDAnalysis.topology.guessers is deprecated in favour of "
    "the new Guessers API. "
    "See MDAnalysis.guesser.default_guesser for more details."
)


@deprecate(release="2.8.0", remove="3.0.0", message=deprecation_msg)
def guess_masses(atom_types):
    """Guess the mass of many atoms based upon their type

    Parameters
    ----------
    atom_types
      Type of each atom

    Returns
    -------
    atom_masses : np.ndarray dtype float64
    """
    validate_atom_types(atom_types)
    masses = np.array(
        [get_atom_mass(atom_t) for atom_t in atom_types], dtype=np.float64
    )
    return masses


@deprecate(release="2.8.0", remove="3.0.0", message=deprecation_msg)
def validate_atom_types(atom_types):
    """Vaildates the atom types based on whether they are available in our tables

    Parameters
    ----------
    atom_types
      Type of each atom

    Returns
    -------
    None

    .. versionchanged:: 0.20.0
       Try uppercase atom type name as well
    """
    for atom_type in np.unique(atom_types):
        try:
            tables.masses[atom_type]
        except KeyError:
            try:
                tables.masses[atom_type.upper()]
            except KeyError:
                warnings.warn(
                    "Failed to guess the mass for the following atom types: {}".format(
                        atom_type
                    )
                )


@deprecate(release="2.8.0", remove="3.0.0", message=deprecation_msg)
def guess_types(atom_names):
    """Guess the atom type of many atoms based on atom name

    Parameters
    ----------
    atom_names
      Name of each atom

    Returns
    -------
    atom_types : np.ndarray dtype object
    """
    return np.array(
        [guess_atom_element(name) for name in atom_names], dtype=object
    )


@deprecate(release="2.8.0", remove="3.0.0", message=deprecation_msg)
def guess_atom_type(atomname):
    """Guess atom type from the name.

    At the moment, this function simply returns the element, as
    guessed by :func:`guess_atom_element`.


    See Also
    --------
    :func:`guess_atom_element`
    :mod:`MDAnalysis.topology.tables`


    """
    return guess_atom_element(atomname)


NUMBERS = re.compile(r"[0-9]")  # match numbers
SYMBOLS = re.compile(r"[*+-]")  # match *, +, -


@deprecate(release="2.8.0", remove="3.0.0", message=deprecation_msg)
def guess_atom_element(atomname):
    """Guess the element of the atom from the name.

    Looks in dict to see if element is found, otherwise it uses the first
    character in the atomname. The table comes from CHARMM and AMBER atom
    types, where the first character is not sufficient to determine the atom
    type. Some GROMOS ions have also been added.

    .. Warning: The translation table is incomplete. This will probably result
                in some mistakes, but it still better than nothing!

    See Also
    --------
    :func:`guess_atom_type`
    :mod:`MDAnalysis.topology.tables`
    """
    if atomname == "":
        return ""
    try:
        return tables.atomelements[atomname.upper()]
    except KeyError:
        # strip symbols
        no_symbols = re.sub(SYMBOLS, "", atomname)

        # split name by numbers
        no_numbers = re.split(NUMBERS, no_symbols)
        no_numbers = list(filter(None, no_numbers))  # remove ''
        # if no_numbers is not empty, use the first element of no_numbers
        name = no_numbers[0].upper() if no_numbers else ""

        # just in case
        if name in tables.atomelements:
            return tables.atomelements[name]

        while name:
            if name in tables.elements:
                return name
            if name[:-1] in tables.elements:
                return name[:-1]
            if name[1:] in tables.elements:
                return name[1:]
            if len(name) <= 2:
                return name[0]
            name = name[:-1]  # probably element is on left not right

        # if it's numbers
        return no_symbols


@deprecate(release="2.8.0", remove="3.0.0", message=deprecation_msg)
def guess_bonds(atoms, coords, box=None, **kwargs):
    r"""Guess if bonds exist between two atoms based on their distance.

    Bond between two atoms is created, if the two atoms are within

    .. math::

          d < f \cdot (R_1 + R_2)

    of each other, where :math:`R_1` and :math:`R_2` are the VdW radii
    of the atoms and :math:`f` is an ad-hoc *fudge_factor*. This is
    the `same algorithm that VMD uses`_.

    Parameters
    ----------
    atoms : AtomGroup
         atoms for which bonds should be guessed
    coords : array
         coordinates of the atoms (i.e., `AtomGroup.positions)`)
    fudge_factor : float, optional
        The factor by which atoms must overlap eachother to be considered a
        bond.  Larger values will increase the number of bonds found. [0.55]
    vdwradii : dict, optional
        To supply custom vdwradii for atoms in the algorithm. Must be a dict
        of format {type:radii}. The default table of van der Waals radii is
        hard-coded as :data:`MDAnalysis.topology.tables.vdwradii`.  Any user
        defined vdwradii passed as an argument will supercede the table
        values. [``None``]
    lower_bound : float, optional
        The minimum bond length. All bonds found shorter than this length will
        be ignored. This is useful for parsing PDB with altloc records where
        atoms with altloc A and B maybe very close together and there should be
        no chemical bond between them. [0.1]
    box : array_like, optional
        Bonds are found using a distance search, if unit cell information is
        given, periodic boundary conditions will be considered in the distance
        search. [``None``]

    Returns
    -------
    list
        List of tuples suitable for use in Universe topology building.

    Warnings
    --------
    No check is done after the bonds are guessed to see if Lewis
    structure is correct. This is wrong and will burn somebody.

    Raises
    ------
    :exc:`ValueError` if inputs are malformed or `vdwradii` data is missing.


    .. _`same algorithm that VMD uses`:
       http://www.ks.uiuc.edu/Research/vmd/vmd-1.9.1/ug/node26.html

    .. versionadded:: 0.7.7
    .. versionchanged:: 0.9.0
       Updated method internally to use more :mod:`numpy`, should work
       faster.  Should also use less memory, previously scaled as
       :math:`O(n^2)`.  *vdwradii* argument now augments table list
       rather than replacing entirely.
    """
    # why not just use atom.positions?
    if len(atoms) != len(coords):
        raise ValueError("'atoms' and 'coord' must be the same length")

    fudge_factor = kwargs.get("fudge_factor", 0.55)

    vdwradii = tables.vdwradii.copy()  # so I don't permanently change it
    user_vdwradii = kwargs.get("vdwradii", None)
    if user_vdwradii:  # this should make algo use their values over defaults
        vdwradii.update(user_vdwradii)

    # Try using types, then elements
    atomtypes = atoms.types

    # check that all types have a defined vdw
    if not all(val in vdwradii for val in set(atomtypes)):
        raise ValueError(
            (
                "vdw radii for types: "
                + ", ".join([t for t in set(atomtypes) if not t in vdwradii])
                + ". These can be defined manually using the"
                + " keyword 'vdwradii'"
            )
        )

    lower_bound = kwargs.get("lower_bound", 0.1)

    if box is not None:
        box = np.asarray(box)

    # to speed up checking, calculate what the largest possible bond
    # atom that would warrant attention.
    # then use this to quickly mask distance results later
    max_vdw = max([vdwradii[t] for t in atomtypes])

    bonds = []

    pairs, dist = distances.self_capped_distance(
        coords, max_cutoff=2.0 * max_vdw, min_cutoff=lower_bound, box=box
    )
    for idx, (i, j) in enumerate(pairs):
        d = (vdwradii[atomtypes[i]] + vdwradii[atomtypes[j]]) * fudge_factor
        if dist[idx] < d:
            bonds.append((atoms[i].index, atoms[j].index))
    return tuple(bonds)


@deprecate(release="2.8.0", remove="3.0.0", message=deprecation_msg)
def guess_angles(bonds):
    """Given a list of Bonds, find all angles that exist between atoms.

    Works by assuming that if atoms 1 & 2 are bonded, and 2 & 3 are bonded,
    then (1,2,3) must be an angle.

    Returns
    -------
    list of tuples
        List of tuples defining the angles.
        Suitable for use in u._topology


    See Also
    --------
    :meth:`guess_bonds`


    .. versionadded 0.9.0
    """
    angles_found = set()

    for b in bonds:
        for atom in b:
            other_a = b.partner(atom)  # who's my friend currently in Bond
            for other_b in atom.bonds:
                if other_b != b:  # if not the same bond I start as
                    third_a = other_b.partner(atom)
                    desc = tuple([other_a.index, atom.index, third_a.index])
                    if desc[0] > desc[-1]:  # first index always less than last
                        desc = desc[::-1]
                    angles_found.add(desc)

    return tuple(angles_found)


@deprecate(release="2.8.0", remove="3.0.0", message=deprecation_msg)
def guess_dihedrals(angles):
    """Given a list of Angles, find all dihedrals that exist between atoms.

    Works by assuming that if (1,2,3) is an angle, and 3 & 4 are bonded,
    then (1,2,3,4) must be a dihedral.

    Returns
    -------
    list of tuples
        List of tuples defining the dihedrals.
        Suitable for use in u._topology

    .. versionadded 0.9.0
    """
    dihedrals_found = set()

    for b in angles:
        a_tup = tuple([a.index for a in b])  # angle as tuple of numbers
        # if searching with b[0], want tuple of (b[2], b[1], b[0], +new)
        # search the first and last atom of each angle
        for atom, prefix in zip(
            [b.atoms[0], b.atoms[-1]], [a_tup[::-1], a_tup]
        ):
            for other_b in atom.bonds:
                if not other_b.partner(atom) in b:
                    third_a = other_b.partner(atom)
                    desc = prefix + (third_a.index,)
                    if desc[0] > desc[-1]:
                        desc = desc[::-1]
                    dihedrals_found.add(desc)

    return tuple(dihedrals_found)


@deprecate(release="2.8.0", remove="3.0.0", message=deprecation_msg)
def guess_improper_dihedrals(angles):
    """Given a list of Angles, find all improper dihedrals that exist between
    atoms.

    Works by assuming that if (1,2,3) is an angle, and 2 & 4 are bonded,
    then (2, 1, 3, 4) must be an improper dihedral.
    ie the improper dihedral is the angle between the planes formed by
    (1, 2, 3) and (1, 3, 4)

    Returns
    -------
        List of tuples defining the improper dihedrals.
        Suitable for use in u._topology

    .. versionadded 0.9.0
    """
    dihedrals_found = set()

    for b in angles:
        atom = b[1]  # select middle atom in angle
        # start of improper tuple
        a_tup = tuple([b[a].index for a in [1, 2, 0]])
        # if searching with b[1], want tuple of (b[1], b[2], b[0], +new)
        # search the first and last atom of each angle
        for other_b in atom.bonds:
            other_atom = other_b.partner(atom)
            # if this atom isn't in the angle I started with
            if not other_atom in b:
                desc = a_tup + (other_atom.index,)
                if desc[0] > desc[-1]:
                    desc = desc[::-1]
                dihedrals_found.add(desc)

    return tuple(dihedrals_found)


@deprecate(release="2.8.0", remove="3.0.0", message=deprecation_msg)
def get_atom_mass(element):
    """Return the atomic mass in u for *element*.

    Masses are looked up in :data:`MDAnalysis.topology.tables.masses`.

    .. Warning:: Unknown masses are set to 0.0

    .. versionchanged:: 0.20.0
       Try uppercase atom type name as well
    """
    try:
        return tables.masses[element]
    except KeyError:
        try:
            return tables.masses[element.upper()]
        except KeyError:
            return 0.0


@deprecate(release="2.8.0", remove="3.0.0", message=deprecation_msg)
def guess_atom_mass(atomname):
    """Guess a mass based on the atom name.

    :func:`guess_atom_element` is used to determine the kind of atom.

    .. warning:: Anything not recognized is simply set to 0; if you rely on the
                 masses you might want to double check.
    """
    return get_atom_mass(guess_atom_element(atomname))


@deprecate(release="2.8.0", remove="3.0.0", message=deprecation_msg)
def guess_atom_charge(atomname):
    """Guess atom charge from the name.

    .. Warning:: Not implemented; simply returns 0.
    """
    # TODO: do something slightly smarter, at least use name/element
    return 0.0


@deprecate(release="2.8.0", remove="3.0.0", message=deprecation_msg)
def guess_aromaticities(atomgroup):
    """Guess aromaticity of atoms using RDKit

    Parameters
    ----------
    atomgroup : mda.core.groups.AtomGroup
        Atoms for which the aromaticity will be guessed

    Returns
    -------
    aromaticities : numpy.ndarray
        Array of boolean values for the aromaticity of each atom


    .. versionadded:: 2.0.0
    """
    mol = atomgroup.convert_to("RDKIT")
    return np.array([atom.GetIsAromatic() for atom in mol.GetAtoms()])


@deprecate(release="2.8.0", remove="3.0.0", message=deprecation_msg)
def guess_gasteiger_charges(atomgroup):
    """Guess Gasteiger partial charges using RDKit

    Parameters
    ----------
    atomgroup : mda.core.groups.AtomGroup
        Atoms for which the charges will be guessed

    Returns
    -------
    charges : numpy.ndarray
        Array of float values representing the charge of each atom


    .. versionadded:: 2.0.0
    """
    mol = atomgroup.convert_to("RDKIT")
    from rdkit.Chem.rdPartialCharges import ComputeGasteigerCharges

    ComputeGasteigerCharges(mol, throwOnParamFailure=True)
    return np.array(
        [atom.GetDoubleProp("_GasteigerCharge") for atom in mol.GetAtoms()],
        dtype=np.float32,
    )



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#

"""
XVG auxiliary reader --- :mod:`MDAnalysis.auxiliary.XVG`
========================================================

xvg files are produced by Gromacs during simulation or analysis, formatted
for plotting data with Grace.

Data is column-formatted; time/data selection is enabled by providing column
indices.

Note
----
By default, the time of each step is assumed to be stored in the first column,
in units of ps.


.. autoclass:: XVGStep
   :members:


XVG Readers
-----------
The default :class:`XVGReader` reads and stores the full contents of the .xvg
file on initialisation, while a second reader (:class:`XVGFileReader`) that
reads steps one at a time as required is also provided for when a lower memory
footprint is desired.

Note
----
Data is assumed to be time-ordered.

Multiple datasets, separated in the .xvg file by '&', are currently not
supported (the readers will stop at the first line starting '&').


.. autoclass:: XVGReader
   :members:

.. autoclass:: XVGFileReader
   :members:


.. autofunction:: uncomment

"""
import numbers
import os
import numpy as np
from . import base
from ..lib.util import anyopen


def uncomment(lines):
    """Remove comments from lines in an .xvg file

    Parameters
    ----------
    lines : list of str
        Lines as directly read from .xvg file

    Yields
    ------
    str
        The next non-comment line, with any trailing comments removed
    """
    for line in lines:
        stripped_line = line.strip()
        # ignore blank lines
        if not stripped_line:
            continue
        # '@' must be at the beginning of a line to be a grace instruction
        if stripped_line[0] == "@":
            continue
        # '#' can be anywhere in the line, everything after is a comment
        comment_position = stripped_line.find("#")
        if comment_position > 0 and stripped_line[:comment_position]:
            yield stripped_line[:comment_position]
        elif comment_position < 0 and stripped_line:
            yield stripped_line
        # if comment_position == 0, then the line is empty


class XVGStep(base.AuxStep):
    """AuxStep class for .xvg file format.

    Extends the base AuxStep class to allow selection of time and
    data-of-interest fields (by column index) from the full set of data read
    each step.

    Parameters
    ----------
    time_selector : int | None, optional
        Index of column in .xvg file storing time, assumed to be in ps. Default
        value is 0 (i.e. first column).
    data_selector : list of int | None, optional
        List of indices of columns in .xvg file containing data of interest to
        be stored in ``data``. Default value is ``None``.
    **kwargs
        Other AuxStep options.

    See Also
    --------
    :class:`~MDAnalysis.auxiliary.base.AuxStep`
    """

    def __init__(self, time_selector=0, data_selector=None, **kwargs):
        super(XVGStep, self).__init__(
            time_selector=time_selector, data_selector=data_selector, **kwargs
        )

    def _select_time(self, key):
        if key is None:
            # here so that None is a valid value; just return
            return
        if isinstance(key, numbers.Integral):
            return self._select_data(key)
        else:
            raise ValueError("Time selector must be single index")

    def _select_data(self, key):
        if key is None:
            # here so that None is a valid value; just return
            return
        if isinstance(key, numbers.Integral):
            try:
                return self._data[key]
            except IndexError:
                errmsg = (
                    f"{key} not a valid index for data with "
                    f"{len(self._data)} columns"
                )
                raise ValueError(errmsg) from None
        else:
            return np.array([self._select_data(i) for i in key])


class XVGReader(base.AuxReader):
    """Auxiliary reader to read data from an .xvg file.

    Default reader for .xvg files. All data from the file will be read and
    stored on initialisation.

    Parameters
    ----------
    filename : str
        Location of the file containing the auxiliary data.
    **kwargs
       Other AuxReader options.

    See Also
    --------
    :class:`~MDAnalysis.auxiliary.base.AuxReader`

    Note
    ----
    The file is assumed to be of a size such that reading and storing the full
    contents is practical.
    """

    format = "XVG"
    _Auxstep = XVGStep

    def __init__(self, filename, **kwargs):
        self._auxdata = os.path.abspath(filename)
        with anyopen(filename) as xvg_file:
            lines = xvg_file.readlines()
        auxdata_values = []
        # remove comments before storing
        for i, line in enumerate(uncomment(lines)):
            if line.lstrip()[0] == "&":
                # multiple data sets not supported; stop at end of first set
                break
            auxdata_values.append([float(val) for val in line.split()])
            # check the number of columns is consistent
            if len(auxdata_values[i]) != len(auxdata_values[0]):
                raise ValueError(
                    "Step {0} has {1} columns instead of "
                    "{2}".format(i, auxdata_values[i], auxdata_values[0])
                )
        self._auxdata_values = np.array(auxdata_values)
        self._n_steps = len(self._auxdata_values)
        super(XVGReader, self).__init__(**kwargs)

    def _memory_usage(self):
        return self._auxdata_values.nbytes

    def _read_next_step(self):
        """Read next auxiliary step and update ``auxstep``.

        Returns
        -------
        AuxStep object
            Updated with the data for the new step.

        Raises
        ------
        StopIteration
            When end of auxiliary data set is reached.
        """
        auxstep = self.auxstep
        new_step = self.step + 1
        if new_step < self.n_steps:
            auxstep._data = self._auxdata_values[new_step]
            auxstep.step = new_step
            return auxstep
        else:
            self.rewind()
            raise StopIteration

    def _go_to_step(self, i):
        """Move to and read i-th auxiliary step.

        Parameters
        ----------
        i: int
            Step number (0-indexed) to move to

        Returns
        -------
        :class:`XVGStep`

        Raises
        ------
        ValueError
            If step index not in valid range.
        """
        if i >= self.n_steps or i < 0:
            raise ValueError(
                "Step index {0} is not valid for auxiliary "
                "(num. steps {1})".format(i, self.n_steps)
            )
        self.auxstep.step = i - 1
        self.next()
        return self.auxstep

    def read_all_times(self):
        """Get NumPy array of time at each step.

        Returns
        -------
        NumPy array of float
            Time at each step.
        """
        return self._auxdata_values[:, self.time_selector]


class XVGFileReader(base.AuxFileReader):
    """Auxiliary reader to read (one step at a time) from an .xvg file.

    An alternative XVG reader which reads each step from the .xvg file as
    needed (rather than reading and storing all from the start), for a lower
    memory footprint.

    Parameters
    ----------
    filename : str
       Location of the file containing the auxiliary data.
    **kwargs
       Other AuxReader options.

    See Also
    --------
    :class:`~MDAnalysis.auxiliary.base.AuxFileReader`


    Note
    ----
    The default reader for .xvg files is :class:`XVGReader`.
    """

    format = "XVG-F"
    _Auxstep = XVGStep

    def __init__(self, filename, **kwargs):
        super(XVGFileReader, self).__init__(filename, **kwargs)

    def _read_next_step(self):
        """Read next recorded step in xvg file and update ``auxstep``.

        Returns
        -------
        AuxStep object
            Updated with the data for the new step.

        Raises
        ------
        StopIteration
            When end of file or end of first data set is reached.
        """
        line = next(self.auxfile)
        while True:
            if not line or (line.strip() and line.strip()[0] == "&"):
                # at end of file or end of first set of data (multiple sets
                # currently not supported)
                self.rewind()
                raise StopIteration
            # uncomment the line
            for uncommented in uncomment([line]):
                # line has data in it; add to auxstep + return
                auxstep = self.auxstep
                auxstep.step = self.step + 1
                auxstep._data = [float(i) for i in uncommented.split()]
                # see if we've set n_cols yet...
                try:
                    auxstep._n_cols
                except AttributeError:
                    # haven't set n_cols yet; set now
                    auxstep._n_cols = len(auxstep._data)
                if len(auxstep._data) != auxstep._n_cols:
                    raise ValueError(
                        f"Step {self.step} has "
                        f"{len(auxstep._data)} columns instead "
                        f"of {auxstep._n_cols}"
                    )
                return auxstep
            # line is comment only - move to next
            line = next(self.auxfile)

    def _count_n_steps(self):
        """Iterate through all steps to count total number.

        Returns
        -------
        int
            Total number of steps
        """
        if not self.constant_dt:
            # check if we've already iterated through to build _times list
            try:
                return len(self._times)
            except AttributeError:
                # might as well build _times now, since we'll need to iterate
                # through anyway
                self._times = self.read_all_times()
                return len(self.read_all_times())
        else:
            # don't need _times; iterate here instead
            self._restart()
            count = 0
            for step in self:
                count = count + 1
            return count

    def read_all_times(self):
        """Iterate through all steps to build times list.

        Returns
        -------
        NumPy array of float
            Time of each step
        """
        self._restart()
        times = []
        for step in self:
            times.append(self.time)
        return np.array(times)



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2.1 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
"""\
=========================================================================
Reading trajectories from memory --- :mod:`MDAnalysis.coordinates.memory`
=========================================================================

:Author: Wouter Boomsma
:Year: 2016
:Copyright: Lesser GNU Public License v2.1+
:Maintainer: Wouter Boomsma <wb@di.ku.dk>, wouterboomsma on github


.. versionadded:: 0.16.0

The module contains a trajectory reader that operates on an array in
memory, rather than reading from file. This makes it possible to
operate on raw coordinates using existing MDAnalysis tools. In
addition, it allows the user to make changes to the coordinates in a
trajectory (e.g. through
:attr:`MDAnalysis.core.groups.AtomGroup.positions`) without having
to write the entire state to file.


How to use the :class:`MemoryReader`
====================================

The :class:`MemoryReader` can be used to either directly generate a
trajectory as a numpy array or by transferring an existing trajectory
to memory.

In-memory representation of arbitrary trajectories
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

If sufficient memory is available to hold a whole trajectory in memory
then analysis can be sped up substantially by transferring the
trajectory to memory.

The most straightforward use of the :class:`MemoryReader` is to simply
use the ``in_memory=True`` flag for the
:class:`~MDAnalysis.core.universe.Universe` class, which
automatically transfers a trajectory to memory::

 import MDAnalysis as mda
 from MDAnalysisTests.datafiles import TPR, XTC

 universe = mda.Universe(TPR, XTC, in_memory=True)

Of course, sufficient memory has to be available to hold the whole
trajectory.


Switching a trajectory to an in-memory representation
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The decision to transfer the trajectory to memory can be made at any
time with the
:meth:`~MDAnalysis.core.universe.Universe.transfer_to_memory` method
of a :class:`~MDAnalysis.core.universe.Universe`::

    universe = mda.Universe(TPR, XTC)
    universe.transfer_to_memory()

This operation may take a while (with `verbose=True` a progress bar is
displayed) but then subsequent operations on the trajectory directly
operate on the in-memory array and will be very fast.


Constructing a Reader from a numpy array
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

The :class:`MemoryReader` provides great flexibility because it
becomes possible to create a :class:`~MDAnalysis.core.universe.Universe` directly
from a numpy array.

A simple example consists of a new universe created from the array
extracted from a DCD
:meth:`~MDAnalysis.coordinates.DCD.DCDReader.timeseries`::

    import MDAnalysis as mda
    from MDAnalysisTests.datafiles import DCD, PSF
    from MDAnalysis.coordinates.memory import MemoryReader

    universe = mda.Universe(PSF, DCD)

    coordinates = universe.trajectory.timeseries(universe.atoms)
    universe2 = mda.Universe(PSF, coordinates, format=MemoryReader, order='afc')


.. _create-in-memory-trajectory-with-AnalysisFromFunction:

.. rubric:: Creating an in-memory trajectory with
            :func:`~MDAnalysis.analysis.base.AnalysisFromFunction`

The :meth:`~MDAnalysis.coordinates.DCD.DCDReader.timeseries` is
currently only implemented for the
:class:`~MDAnalysis.coordinates.DCD.DCDReader`. However, the
:func:`MDAnalysis.analysis.base.AnalysisFromFunction` can provide the
same functionality for any supported trajectory format::

  import MDAnalysis as mda
  from MDAnalysis.tests.datafiles import PDB, XTC

  from MDAnalysis.coordinates.memory import MemoryReader
  from MDAnalysis.analysis.base import AnalysisFromFunction

  u = mda.Universe(PDB, XTC)

  coordinates = AnalysisFromFunction(lambda ag: ag.positions.copy(),
                                     u.atoms).run().results['timeseries']
  u2 = mda.Universe(PDB, coordinates, format=MemoryReader)

.. _creating-in-memory-trajectory-label:

Creating an in-memory trajectory of a sub-system
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Creating a trajectory for just a selection of an existing trajectory
requires the transfer of the appropriate coordinates as well as
creation of a topology of the sub-system. For the latter one can use
the :func:`~MDAnalysis.core.universe.Merge` function, for the former
the :meth:`~MDAnalysis.core.universe.Universe.load_new` method of a
:class:`~MDAnalysis.core.universe.Universe` together with the
:class:`MemoryReader`. In the following, an in-memory trajectory of
only the protein is created::

  import MDAnalysis as mda
  from MDAnalysis.tests.datafiles import PDB, XTC

  from MDAnalysis.coordinates.memory import MemoryReader
  from MDAnalysis.analysis.base import AnalysisFromFunction

  u = mda.Universe(PDB, XTC)
  protein = u.select_atoms("protein")

  coordinates = AnalysisFromFunction(lambda ag: ag.positions.copy(),
                                     protein).run().results['timeseries']
  u2 = mda.Merge(protein)            # create the protein-only Universe
  u2.load_new(coordinates, format=MemoryReader)

The protein coordinates are extracted into ``coordinates`` and then
the in-memory trajectory is loaded from these coordinates. In
principle, this could have all be done in one line::

  u2 = mda.Merge(protein).load_new(
           AnalysisFromFunction(lambda ag: ag.positions.copy(),
                                protein).run().results['timeseries'],
           format=MemoryReader)

The new :class:`~MDAnalysis.core.universe.Universe` ``u2`` can be used
to, for instance, write out a new trajectory or perform fast analysis
on the sub-system.


Classes
=======


.. autoclass:: MemoryReader
   :members:
   :inherited-members:

"""
import logging
import errno
import numpy as np
import warnings
import copy

from . import base
from .timestep import Timestep


# These methods all pass in an existing *view* onto a larger array
def _replace_positions_array(ts, new):
    """Replace the array of positions

    Replaces the array of positions by another array.


    Note
    ----
    The behavior of :meth:`_replace_positions_array` is different from the
    behavior of the :attr:`position` property that replaces the **content**
    of the array. The :meth:`_replace_positions_array` method should only be
    used to set the positions to a different frame in
    :meth:`MemoryReader._read_next_timestep`; there, the memory reader sets
    the positions to a view of the correct frame.  Modifying the positions
    for a given frame should be done with the :attr:`positions` attribute
    that does not break the link between the array of positions in the time
    step and the :attr:`MemoryReader.coordinate_array`.


    .. versionadded:: 0.19.0
    .. versionchanged:: 2.0.0
       This function, and the _repalace helper functions for velocities,
       forces, and dimensions, have been moved out of the now removed
       custom timestep object for :class:`MemoryReader`.
    """
    ts.has_positions = True
    ts._pos = new


def _replace_velocities_array(ts, new):
    ts.has_velocities = True
    ts._velocities = new


def _replace_forces_array(ts, new):
    ts.has_forces = True
    ts._forces = new


def _replace_dimensions(ts, new):
    ts._unitcell = new


class MemoryReader(base.ProtoReader):
    """
    MemoryReader works with trajectories represented as numpy arrays.

    A trajectory reader interface to a numpy array of the coordinates.
    For compatibility with the timeseries interface, support is provided for
    specifying the order of columns through the `order` keyword.

    .. versionadded:: 0.16.0
    .. versionchanged:: 1.0.0
       Support for the deprecated `format` keyword for
       :meth:`MemoryReader.timeseries` has now been removed.
    """

    format = "MEMORY"

    def __init__(
        self,
        coordinate_array,
        order="fac",
        dimensions=None,
        dt=1,
        filename=None,
        velocities=None,
        forces=None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        coordinate_array : numpy.ndarray
            The underlying array of coordinates. The MemoryReader now
            necessarily requires a np.ndarray
        order : {"afc", "acf", "caf", "fac", "fca", "cfa"} (optional)
            the order/shape of the return data array, corresponding
            to (a)tom, (f)rame, (c)oordinates all six combinations
            of 'a', 'f', 'c' are allowed ie "fac" - return array
            where the shape is (frame, number of atoms,
            coordinates).
        dimensions: [A, B, C, alpha, beta, gamma] (optional)
            unitcell dimensions (*A*, *B*, *C*, *alpha*, *beta*, *gamma*)
            lengths *A*, *B*, *C* are in the MDAnalysis length unit (Å), and
            angles are in degrees. An array of dimensions can be given,
            which must then be shape (nframes, 6)
        dt: float (optional)
            The time difference between frames (ps).  If :attr:`time`
            is set, then `dt` will be ignored.
        filename: string (optional)
            The name of the file from which this instance is created. Set to ``None``
            when created from an array
        velocities : numpy.ndarray (optional)
            Atom velocities.  Must match shape of coordinate_array.  Will share order
            with coordinates.
        forces : numpy.ndarray (optional)
            Atom forces.  Must match shape of coordinate_array  Will share order
            with coordinates

        Raises
        ------
        TypeError if the coordinate array passed is not a np.ndarray

        Note
        ----
        At the moment, only a fixed `dimension` is supported, i.e., the same
        unit cell for all frames in `coordinate_array`. See issue `#1041`_.


        .. _`#1041`: https://github.com/MDAnalysis/mdanalysis/issues/1041

        .. versionchanged:: 0.19.0
            The input to the MemoryReader now must be a np.ndarray
            Added optional velocities and forces
        .. versionchanged:: 2.2.0
            Input kwargs are now stored under the :attr:`_kwargs` attribute,
            and are passed on class creation in :meth:`copy`.
        """

        super(MemoryReader, self).__init__()
        self.filename = filename
        self.stored_order = order
        self._kwargs = kwargs

        # See Issue #1685. The block below checks if the coordinate array
        # passed is of shape (N, 3) and if it is, the coordiante array is
        # reshaped to (1, N, 3)
        try:
            if coordinate_array.ndim == 2 and coordinate_array.shape[1] == 3:
                coordinate_array = coordinate_array[np.newaxis, :, :]
        except AttributeError:
            errmsg = (
                "The input has to be a numpy.ndarray that corresponds "
                "to the layout specified by the 'order' keyword."
            )
            raise TypeError(errmsg) from None

        self.set_array(coordinate_array, order)
        self.n_frames = self.coordinate_array.shape[
            self.stored_order.find("f")
        ]
        self.n_atoms = self.coordinate_array.shape[self.stored_order.find("a")]

        if velocities is not None:
            try:
                velocities = np.asarray(velocities, dtype=np.float32)
            except ValueError:
                errmsg = (
                    f"'velocities' must be array-like got "
                    f"{type(velocities)}"
                )
                raise TypeError(errmsg) from None
            # if single frame, make into array of 1 frame
            if velocities.ndim == 2:
                velocities = velocities[np.newaxis, :, :]
            if not velocities.shape == self.coordinate_array.shape:
                raise ValueError(
                    "Velocities has wrong shape {} "
                    "to match coordinates {}"
                    "".format(velocities.shape, self.coordinate_array.shape)
                )
            self.velocity_array = velocities.astype(np.float32, copy=False)
        else:
            self.velocity_array = None

        if forces is not None:
            try:
                forces = np.asarray(forces, dtype=np.float32)
            except ValueError:
                errmsg = f"'forces' must be array like got {type(forces)}"
                raise TypeError(errmsg) from None
            if forces.ndim == 2:
                forces = forces[np.newaxis, :, :]
            if not forces.shape == self.coordinate_array.shape:
                raise ValueError(
                    "Forces has wrong shape {} "
                    "to match coordinates {}"
                    "".format(forces.shape, self.coordinate_array.shape)
                )
            self.force_array = forces.astype(np.float32, copy=False)
        else:
            self.force_array = None

        provided_n_atoms = kwargs.pop("n_atoms", None)
        if provided_n_atoms is not None and provided_n_atoms != self.n_atoms:
            raise ValueError(
                "The provided value for n_atoms ({}) "
                "does not match the shape of the coordinate "
                "array ({})".format(provided_n_atoms, self.n_atoms)
            )

        self.ts = self._Timestep(self.n_atoms, **kwargs)
        self.ts.dt = dt

        if dimensions is None:
            self.dimensions_array = np.zeros(
                (self.n_frames, 6), dtype=np.float32
            )
        else:
            try:
                dimensions = np.asarray(dimensions, dtype=np.float32)
            except ValueError:
                errmsg = (
                    f"'dimensions' must be array-like got "
                    f"{type(dimensions)}"
                )
                raise TypeError(errmsg) from None
            if dimensions.shape == (6,):
                # single box, tile this to trajectory length
                # allows modifying the box of some frames
                dimensions = np.tile(dimensions, (self.n_frames, 1))
            elif dimensions.shape != (self.n_frames, 6):
                raise ValueError(
                    "Provided dimensions array has shape {}. "
                    "This must be a array of shape (6,) or "
                    "(n_frames, 6)".format(dimensions.shape)
                )
            self.dimensions_array = dimensions

        self.ts.frame = -1
        self.ts.time = -1
        self._read_next_timestep()

    @staticmethod
    def _format_hint(thing):
        """For internal use: Check if MemoryReader can operate on *thing*

        .. versionadded:: 1.0.0
        """
        return isinstance(thing, np.ndarray)

    @staticmethod
    def parse_n_atoms(filename, order="fac", **kwargs):
        """Deduce number of atoms in a given array of coordinates

        Parameters
        ----------
        filename : numpy.ndarray
          data which will be used later in MemoryReader
        order : {"afc", "acf", "caf", "fac", "fca", "cfa"} (optional)
            the order/shape of the return data array, corresponding
            to (a)tom, (f)rame, (c)oordinates all six combinations
            of 'a', 'f', 'c' are allowed ie "fac" - return array
            where the shape is (frame, number of atoms,
            coordinates).

        Returns
        -------
        n_atoms : int
          number of atoms in system
        """
        # assume filename is a numpy array
        return filename.shape[order.find("a")]

    def copy(self):
        """Return a copy of this Memory Reader"""
        vels = (
            self.velocity_array.copy()
            if self.velocity_array is not None
            else None
        )
        fors = (
            self.force_array.copy() if self.force_array is not None else None
        )
        dims = self.dimensions_array.copy()

        new = self.__class__(
            self.coordinate_array.copy(),
            order=self.stored_order,
            dimensions=dims,
            velocities=vels,
            forces=fors,
            dt=self.ts.dt,
            filename=self.filename,
            **self._kwargs,
        )
        new[self.ts.frame]

        for auxname, auxread in self._auxs.items():
            new.add_auxiliary(auxname, auxread.copy())
        # since transformations are already applied to the whole trajectory
        # simply copy the property
        new.transformations = self.transformations

        return new

    def set_array(self, coordinate_array, order="fac"):
        """
        Set underlying array in desired column order.

        Parameters
        ----------
        coordinate_array : :class:`~numpy.ndarray` object
            The underlying array of coordinates
        order : {"afc", "acf", "caf", "fac", "fca", "cfa"} (optional)
            the order/shape of the return data array, corresponding
            to (a)tom, (f)rame, (c)oordinates all six combinations
            of 'a', 'f', 'c' are allowed ie "fac" - return array
            where the shape is (frame, number of atoms,
            coordinates).
        """
        # Only make copy if not already in float32 format
        self.coordinate_array = coordinate_array.astype("float32", copy=False)
        self.stored_format = order

    def get_array(self):
        """
        Return underlying array.
        """
        return self.coordinate_array

    def _reopen(self):
        """Reset iteration to first frame"""
        self.ts.frame = -1
        self.ts.time = -1

    def timeseries(
        self, asel=None, atomgroup=None, start=0, stop=-1, step=1, order="afc"
    ):
        """Return a subset of coordinate data for an AtomGroup in desired
        column order. If no selection is given, it will return a view of the
        underlying array, while a copy is returned otherwise.

        Parameters
        ---------
        asel : AtomGroup (optional)
            Atom selection. Defaults to ``None``, in which case the full set of
            coordinate data is returned. Note that in this case, a view
            of the underlying numpy array is returned, while a copy of the
            data is returned whenever `asel` is different from ``None``.

            .. deprecated:: 2.7.0
               asel argument will be renamed to atomgroup in 3.0.0

        atomgroup: AtomGroup (optional)
            Same as `asel`, will replace `asel` in 3.0.0
        start : int (optional)
            the start trajectory frame
        stop : int (optional)
            the end trajectory frame

            .. deprecated:: 2.4.0
               Note that `stop` is currently *inclusive* but will be
               changed in favour of being *exclusive* in version 3.0.

        step : int (optional)
            the number of trajectory frames to skip
        order : {"afc", "acf", "caf", "fac", "fca", "cfa"} (optional)
            the order/shape of the return data array, corresponding
            to (a)tom, (f)rame, (c)oordinates all six combinations
            of 'a', 'f', 'c' are allowed ie "fac" - return array
            where the shape is (frame, number of atoms,
            coordinates).


        .. versionchanged:: 1.0.0
           Deprecated `format` keyword has been removed. Use `order` instead.
        .. versionchanged:: 2.4.0
            ValueError now raised instead of NoDataError for empty input
            AtomGroup
        """
        if asel is not None:
            warnings.warn(
                "asel argument to timeseries will be renamed to"
                "'atomgroup' in 3.0, see #3911",
                category=DeprecationWarning,
            )
            if atomgroup:
                raise ValueError(
                    "Cannot provide both asel and atomgroup kwargs"
                )
            atomgroup = asel

        if stop != -1:
            warnings.warn(
                "MemoryReader.timeseries inclusive `stop` "
                "indexing will be removed in 3.0 in favour of exclusive "
                "indexing",
                category=DeprecationWarning,
            )

        array = self.get_array()
        if order == self.stored_order:
            pass
        elif order[0] == self.stored_order[0]:
            array = np.swapaxes(array, 1, 2)
        elif order[1] == self.stored_order[1]:
            array = np.swapaxes(array, 0, 2)
        elif order[2] == self.stored_order[2]:
            array = np.swapaxes(array, 0, 1)
        elif order[0] == self.stored_order[1]:
            array = np.swapaxes(array, 1, 0)
            array = np.swapaxes(array, 1, 2)
        elif order[0] == self.stored_order[2]:
            array = np.swapaxes(array, 2, 0)
            array = np.swapaxes(array, 1, 2)

        a_index = order.find("a")
        f_index = order.find("f")
        stop_index = stop + 1
        if stop_index == 0:
            stop_index = None
        basic_slice = (
            [slice(None)] * f_index
            + [slice(start, stop_index, step)]
            + [slice(None)] * (2 - f_index)
        )

        # Return a view if either:
        #   1) atomgroup is None
        #   2) atomgroup corresponds to the selection of all atoms.
        array = array[tuple(basic_slice)]
        if atomgroup is None or atomgroup is atomgroup.universe.atoms:
            return array
        else:
            if len(atomgroup) == 0:
                raise ValueError(
                    "Timeseries requires at least one atom " "to analyze"
                )
            # If selection is specified, return a copy
            return array.take(asel.indices, a_index)

    def _read_next_timestep(self, ts=None):
        """copy next frame into timestep"""

        if self.ts.frame >= self.n_frames - 1:
            raise IOError(errno.EIO, "trying to go over trajectory limit")
        if ts is None:
            ts = self.ts
        ts.frame += 1
        f_index = self.stored_order.find("f")
        basic_slice = (
            [slice(None)] * (f_index)
            + [self.ts.frame]
            + [slice(None)] * (2 - f_index)
        )
        _replace_positions_array(ts, self.coordinate_array[tuple(basic_slice)])
        _replace_dimensions(ts, self.dimensions_array[self.ts.frame])
        if self.velocity_array is not None:
            _replace_velocities_array(
                ts, self.velocity_array[tuple(basic_slice)]
            )
        if self.force_array is not None:
            _replace_forces_array(ts, self.force_array[tuple(basic_slice)])

        ts.time = self.ts.frame * self.dt
        return ts

    def _read_frame(self, i):
        """read frame i"""
        # Frame number is incremented to zero by _read_next_timestep()
        self.ts.frame = i - 1
        return self._read_next_timestep()

    def __repr__(self):
        """String representation"""
        return "<{cls} with {nframes} frames of {natoms} atoms>" "".format(
            cls=self.__class__.__name__,
            nframes=self.n_frames,
            natoms=self.n_atoms,
        )

    def add_transformations(self, *transformations):
        """Add all transformations to be applied to the trajectory.

        This function take as list of transformations as an argument. These
        transformations are functions that will be called by the Reader and given
        a :class:`Timestep` object as argument, which will be transformed and returned
        to the Reader.
        The transformations can be part of the :mod:`~MDAnalysis.transformations`
        module, or created by the user, and are stored as a list `transformations`.
        This list can only be modified once, and further calls of this function will
        raise an exception.

        .. code-block:: python

          u = MDAnalysis.Universe(topology, coordinates)
          workflow = [some_transform, another_transform, this_transform]
          u.trajectory.add_transformations(*workflow)

        Parameters
        ----------
        transform_list : list
            list of all the transformations that will be applied to the coordinates

        See Also
        --------
        :mod:`MDAnalysis.transformations`
        """
        # Overrides :meth:`~MDAnalysis.coordinates.base.ProtoReader.add_transformations`
        # to avoid unintended behaviour where the coordinates of each frame are transformed
        # multiple times when iterating over the trajectory.
        # In this method, the trajectory is modified all at once and once only.

        super(MemoryReader, self).add_transformations(*transformations)
        for i, ts in enumerate(self):
            for transform in self.transformations:
                ts = transform(ts)

    def _apply_transformations(self, ts):
        """Applies the transformations to the timestep."""
        # Overrides :meth:`~MDAnalysis.coordinates.base.ProtoReader.add_transformations`
        # to avoid applying the same transformations multiple times on each frame

        return ts



# -*- Mode: python; tab-width: 4; indent-tabs-mode:nil; coding:utf-8 -*-
# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4 fileencoding=utf-8
#
# MDAnalysis --- https://www.mdanalysis.org
# Copyright (c) 2006-2017 The MDAnalysis Development Team and contributors
# (see the file AUTHORS for the full list of names)
#
# Released under the Lesser GNU Public Licence, v2 or any higher version
#
# Please cite your use of MDAnalysis in published work:
#
# R. J. Gowers, M. Linke, J. Barnoud, T. J. E. Reddy, M. N. Melo, S. L. Seyler,
# D. L. Dotson, J. Domanski, S. Buchoux, I. M. Kenney, and O. Beckstein.
# MDAnalysis: A Python package for the rapid analysis of molecular dynamics
# simulations. In S. Benthall and S. Rostrup editors, Proceedings of the 15th
# Python in Science Conference, pages 102-109, Austin, TX, 2016. SciPy.
# doi: 10.25080/majora-629e541a-00e
#
# N. Michaud-Agrawal, E. J. Denning, T. B. Woolf, and O. Beckstein.
# MDAnalysis: A Toolkit for the Analysis of Molecular Dynamics Simulations.
# J. Comput. Chem. 32 (2011), 2319--2327, doi:10.1002/jcc.21787
#
import MDAnalysis as mda
import MDAnalysis.tests.datafiles as datafiles
import numpy as np
import pytest
from MDAnalysis import _GUESSERS, _TOPOLOGY_ATTRS
from MDAnalysis.core.topology import Topology
from MDAnalysis.core.topologyattrs import Atomnames, Atomtypes, Masses
from MDAnalysis.exceptions import NoDataError
from MDAnalysis.guesser.base import GuesserBase, get_guesser
from numpy.testing import assert_allclose, assert_equal


class TestBaseGuesser:

    def test_get_guesser(self):
        class TestGuesser1(GuesserBase):
            context = "test1"

        class TestGuesser2(GuesserBase):
            context = "test2"

        assert get_guesser(TestGuesser1).context == "test1"
        assert get_guesser("test1").context == "test1"
        assert get_guesser(TestGuesser2()).context == "test2"

    def test_get_guesser_with_universe(self):
        class TestGuesser1(GuesserBase):
            context = "test1"

        u = mda.Universe.empty(n_atoms=5)
        guesser = get_guesser(TestGuesser1(), u, foo=1)

        assert len(guesser._universe.atoms) == 5
        assert "foo" in guesser._kwargs

    def test_guess_invalid_attribute(self):
        with pytest.raises(
            ValueError,
            match="default guesser can not guess "
            "the following attribute: foo",
        ):
            mda.Universe(datafiles.PDB_xsmall, to_guess=["foo"])

    def test_guess_attribute_with_missing_parent_attr(self):
        names = Atomnames(np.array(["C", "HB", "HA", "O"], dtype=object))
        masses = Masses(
            np.array([np.nan, np.nan, np.nan, np.nan], dtype=np.float64)
        )
        top = Topology(4, 1, 1, attrs=[names, masses])
        u = mda.Universe(top, to_guess=["masses"])
        assert_allclose(
            u.atoms.masses,
            np.array([12.01100, 1.00800, 1.00800, 15.99900]),
            atol=0,
        )

    def test_force_guessing(self):
        names = Atomnames(np.array(["C", "H", "H", "O"], dtype=object))
        types = Atomtypes(np.array(["1", "2", "3", "4"], dtype=object))
        top = Topology(4, 1, 1, attrs=[names, types])
        u = mda.Universe(top, force_guess=["types"])
        assert_equal(u.atoms.types, ["C", "H", "H", "O"])

    def test_partial_guessing(self):
        types = Atomtypes(np.array(["C", "H", "H", "O"], dtype=object))
        masses = Masses(np.array([0, np.nan, np.nan, 0], dtype=np.float64))
        top = Topology(4, 1, 1, attrs=[types, masses])
        u = mda.Universe(top, to_guess=["masses"])
        assert_allclose(
            u.atoms.masses, np.array([0, 1.00800, 1.00800, 0]), atol=0
        )

    def test_force_guess_priority(self):
        "check that passing the attribute to force_guess have higher power"
        types = Atomtypes(np.array(["C", "H", "H", "O"], dtype=object))
        masses = Masses(np.array([0, np.nan, np.nan, 0], dtype=np.float64))
        top = Topology(4, 1, 1, attrs=[types, masses])
        u = mda.Universe(top, to_guess=["masses"], force_guess=["masses"])
        assert_allclose(
            u.atoms.masses,
            np.array([12.01100, 1.00800, 1.00800, 15.99900]),
            atol=0,
        )

    def test_partial_guess_attr_with_unknown_no_value_label(self):
        "trying to partially guess attribute tha doesn't have declared"
        "no_value_label should gives no effect"
        names = Atomnames(np.array(["C", "H", "H", "O"], dtype=object))
        types = Atomtypes(np.array(["", "", "", ""], dtype=object))
        top = Topology(4, 1, 1, attrs=[names, types])
        u = mda.Universe(top, to_guess=["types"])
        assert_equal(u.atoms.types, ["", "", "", ""])

    def test_guess_topology_objects_existing_read(self):
        u = mda.Universe(datafiles.CONECT)
        assert len(u.atoms.bonds) == 72
        assert list(u.bonds[0].indices) == [623, 630]

        # delete some bonds
        u.delete_bonds(u.atoms.bonds[:10])
        assert len(u.atoms.bonds) == 62
        # first bond has changed
        assert list(u.bonds[0].indices) == [1545, 1552]
        # count number of (1545, 1552) bonds
        ag = u.atoms[[1545, 1552]]
        bonds = ag.bonds.atomgroup_intersection(ag, strict=True)
        assert len(bonds) == 1
        assert not bonds[0].is_guessed

        all_indices = [tuple(x.indices) for x in u.bonds]
        assert (623, 630) not in all_indices

        # test guessing new bonds doesn't remove old ones
        u.guess_TopologyAttrs("default", to_guess=["bonds"])
        assert len(u.atoms.bonds) == 1922
        old_bonds = ag.bonds.atomgroup_intersection(ag, strict=True)
        assert len(old_bonds) == 1
        # test guessing new bonds doesn't duplicate old ones
        assert not old_bonds[0].is_guessed

        new_ag = u.atoms[[623, 630]]
        new_bonds = new_ag.bonds.atomgroup_intersection(new_ag, strict=True)
        assert len(new_bonds) == 1
        assert new_bonds[0].is_guessed

    def test_guess_topology_objects_existing_in_universe(self):
        u = mda.Universe(datafiles.CONECT, to_guess=["bonds"])
        assert len(u.atoms.bonds) == 1922
        assert list(u.bonds[0].indices) == [0, 1]

        # delete some bonds
        u.delete_bonds(u.atoms.bonds[:100])
        assert len(u.atoms.bonds) == 1822
        assert list(u.bonds[0].indices) == [94, 99]

        all_indices = [tuple(x.indices) for x in u.bonds]
        assert (0, 1) not in all_indices

        # guess old bonds back
        u.guess_TopologyAttrs("default", to_guess=["bonds"])
        assert len(u.atoms.bonds) == 1922
        # check TopologyGroup contains new (old) bonds
        assert list(u.bonds[0].indices) == [0, 1]

    def test_guess_topology_objects_force(self):
        u = mda.Universe(datafiles.CONECT, force_guess=["bonds"])
        assert len(u.atoms.bonds) == 1922

        with pytest.raises(NoDataError):
            u.atoms.angles

    def test_guess_topology_objects_out_of_order_init(self):
        u = mda.Universe(
            datafiles.PDB_xsmall,
            to_guess=["dihedrals", "angles", "bonds"],
            guess_bonds=False,
        )
        assert len(u.atoms.angles) == 290
        assert len(u.atoms.dihedrals) == 411

    def test_guess_topology_objects_out_of_order_guess(self):
        u = mda.Universe(datafiles.PDB_xsmall)
        with pytest.raises(NoDataError):
            u.atoms.angles

        u.guess_TopologyAttrs(
            "default", to_guess=["dihedrals", "angles", "bonds"]
        )
        assert len(u.atoms.angles) == 290
        assert len(u.atoms.dihedrals) == 411

    def test_force_guess_overwrites_existing_bonds(self):
        u = mda.Universe(datafiles.CONECT)
        assert len(u.atoms.bonds) == 72

        # This low radius should find no bonds
        vdw = dict.fromkeys(set(u.atoms.types), 0.1)
        u.guess_TopologyAttrs("default", to_guess=["bonds"], vdwradii=vdw)
        assert len(u.atoms.bonds) == 72

        # Now force guess bonds
        u.guess_TopologyAttrs("default", force_guess=["bonds"], vdwradii=vdw)
        assert len(u.atoms.bonds) == 0

    def test_guessing_angles_respects_bond_kwargs(self):
        u = mda.Universe(datafiles.PDB_xsmall)
        assert not hasattr(u.atoms, "angles")

        # This low radius should find no angles
        vdw = dict.fromkeys(set(u.atoms.types), 0.01)

        u.guess_TopologyAttrs("default", to_guess=["angles"], vdwradii=vdw)
        assert len(u.atoms.angles) == 0

        # set higher radii for lots of angles!
        vdw = dict.fromkeys(set(u.atoms.types), 1)
        u.guess_TopologyAttrs("default", force_guess=["angles"], vdwradii=vdw)
        assert len(u.atoms.angles) == 7

    def test_guessing_dihedrals_respects_bond_kwargs(self):
        u = mda.Universe(datafiles.CONECT)
        assert len(u.atoms.bonds) == 72

        u.guess_TopologyAttrs("default", to_guess=["dihedrals"])
        assert len(u.atoms.dihedrals) == 3548
        assert not hasattr(u.atoms, "angles")

    def test_guess_invalid_attribute(self):
        default_guesser = get_guesser("default")
        err = "not a recognized MDAnalysis topology attribute"
        with pytest.raises(KeyError, match=err):
            default_guesser.guess_attr("not_an_attribute")

    def test_guess_unsupported_attribute(self):
        default_guesser = get_guesser("default")
        err = "cannot guess this attribute"
        with pytest.raises(ValueError, match=err):
            default_guesser.guess_attr("tempfactors")

    def test_guess_singular(self):
        default_guesser = get_guesser("default")
        u = mda.Universe(datafiles.PDB_xsmall, to_guess=[])
        assert not hasattr(u.atoms, "masses")

        default_guesser._universe = u
        masses = default_guesser.guess_attr("mass")


def test_Universe_guess_bonds_deprecated():
    with pytest.warns(
        DeprecationWarning, match="`guess_bonds` keyword is deprecated"
    ):
        u = mda.Universe(datafiles.PDB_xsmall, guess_bonds=True)


@pytest.mark.parametrize(
    "universe_input",
    [datafiles.DCD, datafiles.XTC, np.random.rand(3, 3), datafiles.PDB_xsmall],
)
def test_universe_creation_from_coordinates(universe_input):
    mda.Universe(universe_input)


def test_universe_creation_from_specific_array():
    a = np.array(
        [
            [0.0, 0.0, 150.0],
            [0.0, 0.0, 150.0],
            [200.0, 0.0, 150.0],
            [0.0, 0.0, 150.0],
            [100.0, 100.0, 150.0],
            [200.0, 100.0, 150.0],
            [0.0, 200.0, 150.0],
            [100.0, 200.0, 150.0],
            [200.0, 200.0, 150.0],
        ]
    )
    mda.Universe(a, n_atoms=9)
